@@ -29,6 +29,14 @@ class Trace(Info):
     pass
 
 
+class Traces(Info):
+    pass
+
+
+class Value(Info):
+    pass
+
+
 # Exceptions {{{1
 # ParseError {{{2
 class ParseError(Exception):
@@ -101,19 +109,23 @@ class TokenLocation(object):
 
 # Lexer {{{1
 # List of the token names
-reserved = {
-    'HEADER'  : 'HEADER',
-    'TYPE'    : 'TYPE',
-    'SWEEP'   : 'SWEEP',
-    'TRACE'   : 'TRACE',
-    'VALUE'   : 'VALUE',
-    'FLOAT'   : 'FLOAT',
-    'COMPLEX' : 'COMPLEX',
-    'DOUBLE'  : 'DOUBLE',
-    'STRUCT'  : 'STRUCT',
-    'PROP'    : 'PROP',
-    'END'     : 'END',
-}
+reserved = {rw: rw for rw in [
+    'BYTE',
+    'COMPLEX',
+    'DOUBLE',
+    'END',
+    'FLOAT',
+    'GROUP',
+    'HEADER',
+    'INT',
+    'NAN',
+    'PROP',
+    'STRUCT',
+    'SWEEP',
+    'TRACE',
+    'TYPE',
+    'VALUE',
+]}
 tokens = [
     'INTEGER',
     'REAL',
@@ -127,6 +139,7 @@ literals = r'()'
 # Regular expressions that define numbers
 t_INTEGER = r"-?[0-9]+"
 t_REAL = r"[+-]?[0-9]+\.[0-9]*([eE][+-][0-9]+)?"
+t_NAN = r"nan|NaN"
 
 # Regular expression for a string
 t_STRING = r'"[^\\\n"]*"'
@@ -135,7 +148,11 @@ t_STRING = r'"[^\\\n"]*"'
 # Identifiers
 def t_ID(t):
     r'[A-Z]+'
-    t.type = reserved.get(t.value, 'ID')    # check for reserved words
+    t.type = reserved.get(t.value)
+    if t.type is None:
+        loc = TokenLocation(t)
+        t.lexer.skip(1)
+        raise ParseError(f"unknown keyword '{t.value}'.", loc)
     return t
 
 
@@ -156,6 +173,11 @@ def p_contents(p):
     p[0] = (p[1], p[2], p[3], p[4], p[5])
 
 
+def p_contents_without_sweep(p):
+    "contents : header_section type_section value_section end"
+    p[0] = (p[1], p[2], None, None, p[3])
+
+
 def p_header_section(p):
     "header_section : HEADER named_values"
     p[0] = dict(p[2])
@@ -188,7 +210,10 @@ def p_integer_value(p):
 
 
 def p_real_value(p):
-    "value : REAL"
+    """
+    value : REAL
+          | NAN
+    """
     p[0] = float(p[1])
 
 
@@ -244,6 +269,8 @@ def p_kind(p):
         kind : FLOAT
              | DOUBLE
              | COMPLEX
+             | INT
+             | BYTE
              | struct
              | prop
     """
@@ -283,7 +310,33 @@ def p_sweep(p):
 
 def p_trace_section(p):
     "trace_section : TRACE traces"
-    p[0] = p[2]
+    # Partition out the groups from the traces.
+    # A group will have one entry in the list of traces, and it will have a
+    # corresponding entry in the groups directory.  The entry is also a
+    # dictionary that maps the member name to the member type.
+    traces = []
+    groups = {}
+    index = None
+    for trace in p[2]:
+        name = trace.name
+        try:
+            count = int(trace.type)
+            index = 0
+            group_name = name
+            groups[group_name] = {}
+            trace.type = 'GROUP'
+            traces.append(trace)
+            continue
+        except ValueError:
+            pass
+        if index is None:
+            traces.append(trace)
+        else:
+            groups[group_name][trace.name] = trace.type
+            index += 1
+            if index == count:
+                index = None
+    p[0] = (traces, groups)
 
 
 def p_traces(p):
@@ -303,6 +356,13 @@ def p_trace(p):
     p[0] = Trace(name=name, type=type)
 
 
+def p_group_trace(p):
+    "trace : string GROUP INTEGER"
+    name = p[1]
+    count = p[3]
+    p[0] = Trace(name=name, type=count)
+
+
 def p_trace_with_props(p):
     "trace : named_value prop"
     # some psf files place a units property on terminal current traces,
@@ -314,55 +374,92 @@ def p_trace_with_props(p):
 def p_value_section(p):
     "value_section : VALUE values"
     values = {}
-    for n, v in p[2]:
+    for n, t, v in p[2]:
         if n not in values:
-            values[n] = []
-        values[n].append(v)
+            values[n] = Value(type=t, values=[])
+        values[n].values.append(v)
     p[0] = values
 
 
 def p_values(p):
-    "values : values named_signal_value"
+    "values : values signal_value"
     p[1].append(p[2])
     p[0] = p[1]
 
 
 def p_values_last(p):
-    "values : named_signal_value"
+    "values : signal_value"
     p[0] = [p[1]]
 
 
 def p_named_signal_scalar(p):
-    "named_signal_value : string number"
-    p[0] = (p[1], p[2])
+    """
+    signal_value : string numbers
+    """
+    p[0] = (p[1], None, p[2])
 
 
-def p_named_signal_group(p):
-    "named_signal_value : string '(' numbers ')'"
-    p[0] = (p[1], tuple(p[3]))
+def p_named_signal_with_type(p):
+    """
+    signal_value : string string numbers
+    """
+    p[0] = (p[1], p[2], p[3])
 
 
 def p_integer_number(p):
-    "number : INTEGER"
+    "simple_number : INTEGER"
     p[0] = int(p[1])
 
 
-def p_reals_number(p):
-    "number : REAL"
+def p_real_number(p):
+    # props are redundant, so ignore them
+    """
+    simple_number : REAL
+                  | REAL prop
+                  | NAN
+                  | NAN prop
+    """
     p[0] = float(p[1])
 
 
-def p_numbers(p):
-    "numbers : numbers number"
+def p_simple_numbers(p):
+    "simple_numbers : simple_numbers simple_number"
     p[1].append(p[2])
     p[0] = p[1]
 
 
-def p_numbers_last(p):
-    "numbers : number"
+def p_simple_numbers_last(p):
+    "simple_numbers : simple_number"
     p[0] = [p[1]]
 
 
+def p_composite_number(p):
+    """
+    composite_number : '(' simple_numbers ')'
+                     | '(' simple_numbers ')' prop
+    """
+    p[0] = tuple(p[2])
+
+
+def p_composite_numbers(p):
+    "composite_numbers : composite_numbers composite_number"
+    p[1].append(p[2])
+    p[0] = p[1]
+
+
+def p_composite_numbers_last(p):
+    "composite_numbers : composite_number"
+    p[0] = [p[1]]
+
+
+def p_numbers(p):
+    """
+    numbers : simple_numbers
+            | composite_numbers
+    """
+    p[0] = p[1]
+
+
 def p_end(p):
     "end : END"
 
